#!/bin/sh 
# the next line restarts using wish \
exec wish "$0" "$@"
# ########################################################
#
# LED - Cube - Editor
#
# 20070617
#
# Ziele:
# 1. Daten für LED-Cube von Hand erstellen / bearbeiten,
#    ... Datenimport / -konvertierung von Bildern / Icons,
#    ... Helligkeitseingabe über Tasten z.B. 0..9A..F
# 2. Bewegungsfunktionen: Bewegen, Drehen
#    ... ? Drehen ... zus. Nachbarpunkte im Raum einschalten
# 3. Bildfolgen in einer einfachen Playliste zusammenstellen,
#    zeitgesteuert an den Server weiterleiten
# 4. ... Direkte Auswahl im Anzeigefenster
#
# Tastenbelegungen:
#  <KeyRelease-Left> ............. Auswahl in -X
#  <KeyRelease-Right> ............ Auswahl in +X
#  <KeyRelease-Up> ............... Auswahl in -Y
#  <KeyRelease-Down> ............. Auswahl in +Y
#  <KeyRelease-Next> ............. Auswahl folg. Ebene in -Z
#  <KeyRelease-Prior> ............ Auswahl folg. Ebene in +Z
#  <Shift-Left> .................. Schieben in -X Ebene/Cube
#  <Shift-Right> ................. Schieben in +X Ebene/Cube
#  <Shift-Up> .................... Schieben in -Y Ebene/Cube
#  <Shift-Down> .................. Schieben in +Y Ebene/Cube
#  <KeyRelease-Return> <Space> ... akt. LED an/aus Ebene/Cube
#  <Control-c> <Control-Insert> .. internes Copy der Ebene
#  <Control-x> <Shift-Delete> .... internes Cut der Ebene
#  <Control-v> <Shift-Insert> .... internes Paste Ebene/Cube
#  <Delete> ...................... Alle LED aus Ebene/Cube
#  <Insert> ...................... Alle LED an Ebene/Cube
#  <Control-s> ................... Bild senden
#  <Control-G> ................... LED-Cube Viewer schreiben
#
# ####################################################### #

global busy_flag
set busy_flag 0

global workdir inifilename
set workdir [ file dirname $argv0 ]
cd $workdir
set workdir [ pwd ]
set inifilename "led_cube_edit.ini"

global img
set img {}

global blocksize blocksize_alt buttons
set blocksize 30
set blocksize_alt 0

global cube level line

source client_lib.tcl
# set gvar(host) "localhost"
# set gvar(port) 12345
# set gvar(led_count) 0
# set gvar(update_refresh) 0
#
# proc receive {sock}
# proc update_ok {}
# proc set_xyz {x y z}
# proc init_led {}
# proc send_led {}
# proc send_cmd {cmd}
# proc draw_point {x y z}
# proc get_distance {x1 y1 z1 x2 y2 z2}
# proc draw_line {x1 y1 z1 x2 y2 z2}
# proc connection_start {}

source transformationsmatrix.tcl
# proc pi {}
# proc matrix_unit { }
# proc matrix_set { matrix }
# proc matrix_scale { matrix a b c }
# proc matrix_move { matrix x y z }
# proc matrix_x_rotate { matrix phi }
# proc matrix_y_rotate { matrix phi }
# proc matrix_z_rotate { matrix phi }
# proc matrix_transform { matrix tmatrix }
# proc vector_value { vector }
# proc vector_diff { vector1 vector2 }
# proc vector_scalarproduct { vector1 vector2 }
# proc vector_angle { vector1 vector2 }
# proc vectors_bounds { vectors bounds }
# proc vectors_transform { vectors matrix }

# ############################################################################
#global cube level line
proc cube_init { { cm { 8 8 8 } } } {
  global cube level line
  set cube(max) $cm
  if { ! [ info exists "cube(xmax_old)" ] } then {
    set cube(xmax_old) 0
    set cube(ymax_old) 0
    set cube(zmax_old) 0
    set cube(x_angle) 90
    set cube(y_angle) 90
    set cube(z_angle) 90
  }
  set cube(xmax) [lindex $cube(max) 0]
  set cube(ymax) [lindex $cube(max) 1]
  set cube(zmax) [lindex $cube(max) 2]
  set cube(x) 0
  set cube(y) 0
  set cube(z) 0
  set cube(0) {}
  set level(0) {}
  set line(0) {}
  for { set x 0 } { $x < $cube(xmax) } { incr x } { lappend line(0) 0 }
  for { set y 0 } { $y < $cube(ymax) } { incr y } { lappend level(0) $line(0) }
  for { set z 0 } { $z < $cube(zmax) } { incr z } { lappend cube(0) $level(0) }
  set cube(1) $cube(0)
  set level(1) [ lindex $cube(1) $cube(z) ]
  set line(1) [ lindex $level(1) $cube(y) ]
  set level(2) $level(0)
  set cube(action) 0
  set cube(rotate) 0
  set cube(xyz) {}
  set cube(vectors) {}
  set cube(matrix) {}
  set cube(script) {}
}

proc cube_load_file { filename } {
  global cube
  if { [ file exists $filename ] } then {
    if { [ catch {open [file join [file nativename $filename]] r} fid ] } {
      puts stderr "Could not open for reading\n$fid"
    } else {
      set max 0
      while { ! [ eof $fid ] } {
        set l [ gets $fid ]
        if { [ llength $l ] > 1 } then {
          if { $max } then {
            set cube(1) $l
            start
            close $fid
            return
          } else {
            set cube(max) $l
            cube_init $cube(max)
            set max 1
          }
#puts $l
        }
      }
      close $fid
    }
  } else {
    puts stderr "Could not find file for reading\n$filename"
  }
}

proc cube_save_file { filename } {
  global cube
  if { [ catch {open [file join [file nativename "$filename"]] w} fid ] } {
    puts stderr "Could not open for writing\n$fid"
  } else {
    puts $fid $cube(max)
    puts $fid $cube(1)
    close $fid
  }
}

proc cube_load { } {
  global workdir cube

  set types {
    { { Led-Cube Files } {.led_cube} }
    { { Text Files }     {.txt}      }
    { { All Files }      *           }
  }
  set filename [tk_getOpenFile -parent . -title "Select a file to load" \
     -initialdir "$workdir" -filetypes $types ]
  cube_load_file $filename
}

proc cube_save { } {
  global workdir cube

  set types {
    { { Led-Cube Files } {.led_cube} }
    { { Text Files }     {.txt}      }
    { { All Files }      *           }
  }
  set filename [tk_getSaveFile -parent . -title "Select a file to save" \
     -initialdir "$workdir" -filetypes $types -defaultextension ".led_cube" ]
  cube_save_file $filename
}

proc cube_generate_3dv { } {
  global workdir cube
  set filename "$workdir/cube_${cube(xmax)}x${cube(ymax)}x${cube(zmax)}_gen.3dv"

  if { [ catch {open [file join [file nativename "$filename"]] w} fid ] } {
    puts stderr "Could not open for writing\n$fid"
  } else {
    puts $fid "set lvtx {"
    for { set z 0 } { $z < $cube(zmax) } { incr z } {
      for { set y 0 } { $y < $cube(ymax) } { incr y } {
        for { set x 0 } { $x < $cube(xmax) } { incr x } {
          set cx [ expr 0 - ($cube(xmax) - 1.0) / $cube(xmax) + $x * 2.0 / $cube(xmax) ]
          set cy [ expr 0 - ($cube(ymax) - 1.0) / $cube(ymax) + $y * 2.0 / $cube(ymax) ]
          set cz [ expr 0 - ($cube(zmax) - 1.0) / $cube(zmax) + $z * 2.0 / $cube(zmax) ]
          puts $fid "{ $cx $cy $cz }"
        }
      }
    }
    puts $fid "}"
    puts $fid "set lcnx {"
    for { set z 0 } { $z < $cube(zmax) } { incr z } {
      for { set y 0 } { $y < $cube(ymax) } { incr y } {
        for { set x 0 } { $x < $cube(xmax) } { incr x } {
          set cx [ expr $x - 1 ]
          set cy [ expr $y - 1 ]
          set cz [ expr $z - 1 ]
          set cxyz1 [ expr $x + $y * $cube(xmax) + $z * $cube(xmax) * $cube(ymax) ]
          if { $cx >= 0 } then {
            set cxyz2 [ expr $cx + $y * $cube(xmax) + $z * $cube(xmax) * $cube(ymax) ]
            puts $fid "{ $cxyz1 $cxyz2 }"
          }
          if { $cy >= 0 } then {
            set cxyz2 [ expr $x + $cy * $cube(xmax) + $z * $cube(xmax) * $cube(ymax) ]
            puts $fid "{ $cxyz1 $cxyz2 }"
          }
          if { $cz >= 0 } then {
            set cxyz2 [ expr $x + $y * $cube(xmax) + $cz * $cube(xmax) * $cube(ymax) ]
            puts $fid "{ $cxyz1 $cxyz2 }"
          }
        }
      }
    }
    puts $fid "}"
    puts $fid "set ldx ${cube(xmax)}"
    puts $fid "set ldy ${cube(ymax)}"
    puts $fid "set ldz ${cube(zmax)}"
    close $fid
    puts "cube_generate_3dv : \'${filename}\'"
  }
}

proc cube_set_1 { } { # ok
  global cube level line
  if { $cube(x) < 0 } then { set cube(x) 0 }
  if { $cube(x) >= $cube(xmax) } then { set cube(x) [ expr $cube(xmax) - 1 ] }
  if { $cube(y) < 0 } then { set cube(y) 0 }
  if { $cube(y) >= $cube(ymax) } then { set cube(y) [ expr $cube(ymax) - 1 ] }
  if { $cube(z) < 0 } then { set cube(z) 0 }
  if { $cube(z) >= $cube(zmax) } then { set cube(z) [ expr $cube(zmax) - 1 ] }
  set level(1) [ lindex $cube(1) $cube(z) ]
  set line(1) [ lindex $level(1) $cube(y) ]
}

proc cube_set_led { x y z ledon } { # ok
  global cube level line

  if { $cube(action) } then {
    set c {}
    foreach lev $cube(1) {
      set lin [ lreplace [lindex $lev $y] $x $x $ledon ]
      lappend c [ lreplace $lev $y $y $lin ]
    }
    set cube(1) $c
    set level(1) [ lindex $cube(1) $cube(z) ]
    set line(1) [ lindex $level(1) $cube(y) ]
  } else {
    set line(1) [ lreplace $line(1) $x $x $ledon ]
    set level(1) [ lreplace $level(1) $y $y $line(1) ]
    set cube(1) [ lreplace $cube(1) $z $z $level(1) ]
  }
  graf_refresh
}

proc cube_level_0 { } { # ok
  global cube level
  global buttons busy_flag

  if { ! $busy_flag } then {
    set level(1) $level(0)
    if { $cube(action) } then {
      set cube(1) $cube(0)
    } else {
      set cube(1) [ lreplace $cube(1) $cube(z) $cube(z) $level(1) ]
    }
    buttons_reset
    set busy_flag 0
  }
  focus ${buttons}.but($cube(x),$cube(y))
}

proc cube_level_1 { { led 1 } } { # ok
  global cube level
  global buttons busy_flag

  if { ! $busy_flag } then {
    set p {}
    foreach lin [ lindex $level(0) 0 ] { lappend p $led }
    set level(1) {}
    foreach lin $level(0) { lappend level(1) $p }
    if { $cube(action) } then {
      for { set z 0 } { $z < $cube(zmax) } { incr z 1 } {
        set cube(1) [ lreplace $cube(1) $z $z $level(1) ]
      }
    } else {
      set cube(1) [ lreplace $cube(1) $cube(z) $cube(z) $level(1) ]
    }
    buttons_reset
    set busy_flag 0
  }
  focus ${buttons}.but($cube(x),$cube(y))
}

proc cube_level_copy { } { # ok
  global cube level
  set level(2) $level(1)
}

proc cube_level_cut { } { # ok
  global cube level
  global buttons busy_flag

  if { ! $busy_flag } then {
    set level(2) $level(1)
    set level(1) $level(0)
    set cube(1) [ lreplace $cube(1) $cube(z) $cube(z) $level(1) ]
    buttons_reset
    set busy_flag 0
  }
  focus ${buttons}.but($cube(x),$cube(y))
}

proc cube_level_paste { } { # ok
  global cube level
  global buttons busy_flag

  if { ! $busy_flag } then {
    set level(1) $level(2)
    if { $cube(action) } then {
      for { set z 0 } { $z < $cube(zmax) } { incr z 1 } {
        set cube(1) [ lreplace $cube(1) $z $z $level(1) ]
      }
    } else {
      set cube(1) [ lreplace $cube(1) $cube(z) $cube(z) $level(1) ]
    }
    buttons_reset
    set busy_flag 0
  }
  focus ${buttons}.but($cube(x),$cube(y))
}

proc cube_level_move { dx dy z } { # ok
  global cube level line
  if { $dy } then {
    if { $dy > 0 } then {
      set level(1) [ lrange $level(1) 0 [expr $cube(ymax) - 2] ]
      set level(1) [ linsert $level(1) 0 $line(0) ]
    } else {
      set level(1) [ lrange $level(1) 1 [expr $cube(ymax) - 1] ]
      lappend level(1) $line(0)
    }
  }
  if { $dx } then {
    for { set y 0 } { $y < $cube(ymax) } { incr y 1 } {
      set lin [ lindex $level(1) $y ]
      if { $dx > 0 } then {
        set lin [ lrange $lin 0 [expr $cube(xmax) - 2] ]
        set lin [ linsert $lin 0 0 ]
      } else {
        set lin [ lrange $lin 1 [expr $cube(xmax) - 1] ]
        lappend lin 0
      }
      set level(1) [ lreplace $level(1) $y $y $lin ]
    }
  }
  if { ! $cube(action) } then {
    set line(1) [ lindex $level(1) $cube(y) ]
  }
  set cube(1) [ lreplace $cube(1) $z $z $level(1) ]
}

proc cube_move { dx dy dz } { # ok
  global cube level line
  if { $dx || $dy } then {
    for { set z 0 } { $z < $cube(zmax) } { incr z 1 } {
      set level(1) [ lindex $cube(1) $z ]
      cube_level_move $dx $dy $z
    }
  }
  if { $dz } then {
    if { $dz > 0 } then {
      set cube(1) [ lrange $cube(1) 0 [expr $cube(zmax) - 2] ]
      set cube(1) [ linsert $cube(1) 0 $level(0) ]
    } else {
      set cube(1) [ lrange $cube(1) 1 [expr $cube(zmax) - 1] ]
      lappend cube(1) $level(0)
    }
  }
  set level(1) [ lindex $cube(1) $cube(z) ]
  set line(1) [ lindex $level(1) $cube(y) ]
}

proc cube_switch_rotate { } {
  global cube
  cube_mode_edit_or_xyz $cube(rotate)
}

proc cube_mode_edit_or_xyz { mode } {
  global cube

  if { $mode == 0 } then {
# auskommentiert, um Aenderungen an der Rotationsform
# vornehman zu koennen:
#    set cube(1) $cube(xyz)
    set cube(vectors) {}
    set cube(matrix) {}
    set cube(xyz) $cube(0)
  } else {
    set cube(xyz) $cube(1)
    set cube(vectors) [ cube_to_vectors ]
    set cube(matrix) [ matrix_unit ]
  }
#puts "mode $mode"
#puts "cube(rotate) $cube(rotate)"
#puts "cube(vectors) $cube(vectors)"
#puts "cube(matrix) $cube(matrix)"
}

proc cube_level_rotate { dx dy z } { # offen
  global cube level line

  set cube(matrix) [ matrix_move $cube(matrix) [expr 0.0 - ($cube(xmax)-1) / 2.0] \
    [expr 0.0 - ($cube(ymax)-1) / 2.0] [expr 0.0 - ($cube(zmax)-1) / 2.0] ]
  if { $dx } then {
    set cube(matrix) [ matrix_z_rotate $cube(matrix) [expr 0 - $dx * $cube(z_angle)] ]
  }
  if { $dy } then {
    set cube(matrix) [ matrix_z_rotate $cube(matrix) [expr 0 - $dy * $cube(z_angle)] ]
  }
  set cube(matrix) [ matrix_move $cube(matrix) [expr ($cube(xmax)-1) / 2.0] \
    [expr ($cube(ymax)-1) / 2.0] [expr ($cube(zmax)-1) / 2.0] ]
  cube_from_vectors [ vectors_transform $cube(vectors) $cube(matrix) ]
  set level(1) [ lindex $cube(1) $cube(z) ]
  set line(1) [ lindex $level(1) $cube(y) ]
#  if { ! $cube(action) } then {
#    set line(1) [ lindex $level(1) $cube(y) ]
#  }
#  set cube(1) [ lreplace $cube(1) $z $z $level(1) ]
}

proc cube_rotate { dx dy dz } { # offen
  global cube level line
  set cube(matrix) [ matrix_move $cube(matrix) [expr 0.0 - ($cube(xmax)-1) / 2.0] \
    [expr 0.0 - ($cube(ymax)-1) / 2.0] [expr 0.0 - ($cube(zmax)-1) / 2.0] ]
  if { $dx } then {
    set cube(matrix) [ matrix_y_rotate $cube(matrix) [expr $dx * $cube(y_angle)] ]
  }
  if { $dy } then {
    set cube(matrix) [ matrix_x_rotate $cube(matrix) [expr $dy * $cube(x_angle)] ]
  }
  if { $dz } then {
    set cube(matrix) [ matrix_z_rotate $cube(matrix) [expr $dz * $cube(z_angle)] ]
  }
  set cube(matrix) [ matrix_move $cube(matrix) [expr ($cube(xmax)-1) / 2.0] \
    [expr ($cube(ymax)-1) / 2.0] [expr ($cube(zmax)-1) / 2.0] ]
  cube_from_vectors [ vectors_transform $cube(vectors) $cube(matrix) ]
  set level(1) [ lindex $cube(1) $cube(z) ]
  set line(1) [ lindex $level(1) $cube(y) ]
}

proc cube_trans { dx dy dz } { # ok
  global cube
  global blocksize buttons
  global busy_flag

  if { ! $busy_flag } then {
    set busy_flag 1
    if { $cube(rotate) } then {
      if { $cube(action) } then {
        cube_rotate $dx $dy $dz
      } else {
        cube_level_rotate $dx $dy $cube(z)
      }
    } else {
      if { $cube(action) } then {
        cube_move $dx $dy $dz
      } else {
        cube_level_move $dx $dy $cube(z)
      }
    }
    buttons_reset
    set busy_flag 0
  }
  focus ${buttons}.but($cube(x),$cube(y))
}

proc cube_to_vectors { } {
  global cube

  set vectors {}
  for { set z 0 } { $z <= $cube(zmax) } { incr z 1 } {
    set y 0
    foreach lin [ lindex $cube(1) $z ] {
      set x 0
      foreach led $lin {
        if { $led > 0 } then {
          lappend vectors [ list $x $y $z ]
        }
        incr x 1
      }
      incr y 1
    }
  }
  return $vectors
}

proc cube_from_vectors { vectors } {
  global cube
  set regexp_vector_3 {^([+-\.0-9A-Za-z]*)[ ]*([+-\.0-9A-Za-z]*)[ ]*([+-\.0-9A-Za-z]*)$}
  set regexp_vector_4 {^([+-\.0-9A-Za-z]*)[ ]*([+-\.0-9A-Za-z]*)[ ]*([+-\.0-9A-Za-z]*)[ ]*([+-\.0-9A-Za-z]*)$}

  set cube(1) $cube(0)
  foreach vector $vectors {
    if { [ llength $vector ] == 4 } then {
      regexp $regexp_vector_4 [join $vector " "] xyz x y z led
    } else {
      set led 1
      regexp $regexp_vector_3 [join $vector " "] xyz x y z
    }
    set x [ expr round($x) ]
    set y [ expr round($y) ]
    set z [ expr round($z) ]
    if { ($x>=0) && ($x<$cube(xmax)) \
      && ($y>=0) && ($y<$cube(ymax)) \
      && ($z>=0) && ($z<$cube(zmax)) } then {
      set lev [ lindex $cube(1) $z ]
      set lin [ lindex $lev $y ]
      set lin [ lreplace $lin $x $x $led ]
      set lev [ lreplace $lev $y $y $lin ]
      set cube(1) [ lreplace $cube(1) $z $z $lev ]
    }
  }
}

proc cube_vectors_send { vectors } {
  global cube gvar

  if { ! [ info exists "gvar(x)" ] } then {
    connection_start
    if { ! [ info exists "gvar(x)" ] } then {
      puts stderr "Server nicht initialisiert!"
      return
    }
  }

  set regexp_vector_3 {^([+-\.0-9A-Za-z]*)[ ]*([+-\.0-9A-Za-z]*)[ ]*([+-\.0-9A-Za-z]*)$}
  set regexp_vector_4 {^([+-\.0-9A-Za-z]*)[ ]*([+-\.0-9A-Za-z]*)[ ]*([+-\.0-9A-Za-z]*)[ ]*([+-\.0-9A-Za-z]*)$}

  if { ($cube(xmax)==$gvar(x)) && ($cube(ymax)==$gvar(y)) && ($cube(zmax)==$gvar(z)) } then {
    init_led
    foreach vector $vectors {
      if { [ llength $vector ] == 4 } then {
        regexp $regexp_vector_4 [join $vector " "] xyz x y z led
      } else {
        set led 1
        regexp $regexp_vector_3 [join $vector " "] xyz x y z
      }
      set x [ expr round($x) ]
      set y [ expr round($y) ]
      set z [ expr round($z) ]
      if { ($x>=0) && ($x<$cube(xmax)) \
        && ($y>=0) && ($y<$cube(ymax)) \
        && ($z>=0) && ($z<$cube(zmax)) } then {
#        draw_point $x $y $z $led
        draw_point $x $y $z
      }
    }
    send_led
  } else {
    puts stderr "Dimension falsch! Cube ${cube(xmax)}x${cube(ymax)}x${cube(zmax)} ${gvar(x)}x${gvar(y)}x${gvar(z)}"
  }
}

proc cube_send { } {
  global cube gvar

  if { ! [ info exists "gvar(x)" ] } then {
    connection_start
    if { ! [ info exists "gvar(x)" ] } then {
      puts stderr "Server nicht initialisiert!"
      return
    }
  }

  if { ($cube(xmax)==$gvar(x)) && ($cube(ymax)==$gvar(y)) && ($cube(zmax)==$gvar(z)) } then {
    init_led
    for { set z 0 } { $z <= $cube(zmax) } { incr z 1 } {
      set y 0
      foreach lin [ lindex $cube(1) $z ] {
        set x 0
        foreach led $lin {
          if { $led > 0 } then {
#        draw_point $x $y $z $led
            draw_point $x $y $z
          }
          incr x 1
        }
        incr y 1
      }
    }
    send_led
  } else {
    puts stderr "Dimension falsch! Cube ${cube(xmax)}x${cube(ymax)}x${cube(zmax)} ${gvar(x)}x${gvar(y)}x${gvar(z)}"
  }
}

# ############################################################################
proc drawled { win x y blocksize c1 c2 { tag "temp" } } {
  $win create oval [ expr $x + 1 ] [ expr $y + 1 ] \
    [ expr $x + $blocksize - 1 ] [ expr $y + $blocksize - 1 ] \
     -fill $c1 -width 1 -outline $c2 -tags $tag
}

proc drawtext { win x y a c1 t { tag "temp" } } {
  global fnt

  if { [ string compare $a "e" ] } then {
    set j "right"
  } else {
    set j "left"
  }
  $win create text $x $y -anchor n$a -justify $j -fill $c1 -text "$t" -font "$fnt" -tags $tag
}

# ############################################################################
proc graf_create { win } {
  frame $win -borderwidth 2 -relief sunken -class graf
  canvas $win.graf -borderwidth 0 -background white \
    -highlightthickness 0
  pack $win.graf -expand yes -fill both

  bind $win.graf <Configure> "graf_newsize $win.graf; graf_draw $win.graf"

  graf_newsize $win.graf
  return $win
}

proc graf_refresh { } {
  graf_draw .fr.screen.graf
}

proc graf_reset { } {
  graf_refresh
}

proc graf_size { width height } {
  .fr.screen.graf configure -width $width -height $height
}

proc graf_newsize { win } {
  global blocksize blocksize_alt fnt
  global width1 height1

  set width1 [ winfo width $win ]
  set height1 [ winfo height $win ]

  if { $blocksize != $blocksize_alt } then {
# ...
    set blocksize_alt $blocksize
  }

  $win delete all
  $win create rectangle 0 0 $width1 $height1 -fill "#080808" -width 2 -outline "#080808"
}

proc graf_draw { win } {
  global blocksize cube

  $win delete "temp"

  for { set z 0 } { $z <= $cube(z) } { incr z 1 } {
    set ledsize [ expr round( $blocksize * ( 1.0 - ( $cube(z) - $z ) / 15.0 ) )  ]
    set ledshift [ expr round( $blocksize * ( 1.0 - ( $cube(z) - $z ) / 10.0 ) )  ]
#puts "blocksize $blocksize cube(z) $cube(z) z $z ledshift $ledshift"
    set y 0
    foreach lin [ lindex $cube(1) $z ] {
      set x 0
      foreach led $lin {
        if { $led > 0 } then { set c1 "#c0c000" } else { set c1 "#404000" }
        drawled $win [ expr 2+($x+2.5)*1.5*$blocksize-2*$ledshift ] [ expr 2+($cube(ymax)-1-$y+2.5)*1.5*$blocksize-2*$ledshift ] $ledsize $c1 "#202000"
        incr x 1
      }
      incr y 1
    }
  }
}

# ############################################################################
proc focus_move { dx dy dz { refresh -9999 } } {
  global cube level line
  global blocksize buttons
  global busy_flag

  if { ! $busy_flag } then {
    set busy_flag 1
    set max 0
    if { $dz } then {
      set max $cube(zmax)
      incr cube(z) $dz
      if { $cube(z)<0 } then { set cube(z) [expr $max - 1] }
      if { $cube(z)>=$max } then { set cube(z) 0 }
      set level(1) [ lindex $cube(1) $cube(z) ]
    }
    if { $dz || ($refresh > -9999) } then { buttons_reset }
    if { $dy } then {
      set max $cube(ymax)
      incr cube(y) $dy
      if { $cube(y)<0 } then { set cube(y) [expr $max - 1] }
      if { $cube(y)>=$max } then { set cube(y) 0 }
      set line(1) [ lindex $level(1) $cube(y) ]
    }
    if { $dx } then {
      set max $cube(xmax)
      incr cube(x) $dx
      if { $cube(x)<0 } then { set cube(x) [expr $max - 1] }
      if { $cube(x)>=$max } then { set cube(x) 0 }
    }
    ${buttons} delete "temp"
    for { set i 0 } { $i < $cube(z) } { incr i } {
      set x1 [ expr 2+($blocksize+4)*1+2*($cube(z)-1-$i) ]
      set x2 [ expr 2+($blocksize+4)*($cube(xmax)+1)+2*($cube(z)-1-$i) ]
      set y1 [ expr 2+($blocksize+4)*1+2*($cube(z)-1-$i) ]
      set y2 [ expr 2+($blocksize+4)*($cube(ymax)+1)+2*($cube(z)-1-$i) ]
      ${buttons} create line $x2 $y1 $x2 $y2 -width 1 -tags "temp"
      ${buttons} create line $x1 $y2 $x2 $y2 -width 1 -tags "temp"
    }
    for { set i [expr $cube(z) + 1] } { $i < $cube(zmax) } { incr i } {
      set x1 [ expr 2+($blocksize+4)*1-2*($i-$cube(z)) ]
      set x2 [ expr 2+($blocksize+4)*($cube(xmax)+1)-2*($i-$cube(z)-1) ]
      set y1 [ expr 2+($blocksize+4)*1-2*($i-$cube(z)) ]
      set y2 [ expr 2+($blocksize+4)*($cube(ymax)+1)-2*($i-$cube(z)-1) ]
      ${buttons} create line $x1 $y1 $x1 $y2 -width 1 -tags "temp"
      ${buttons} create line $x1 $y1 $x2 $y1 -width 1 -tags "temp"
    }
    focus ${buttons}.but($cube(x),$cube(y))
#puts "focus_move $dx $dy $dz $max -> $cube(x) $cube(y) $cube(z)"
    set busy_flag 0
  }
}

proc buttons_reset { { build {} } } {
  global blocksize buttons img level

  cube_set_1
  set y 0
  foreach lin $level(1) {
    set x 0
    foreach p $lin {
      if { $p } then { set img_ [ lindex $img 1 ] } else { set img_ [ lindex $img 0 ] }
      ${buttons}.but($x,$y) configure -image $img_
      incr x
    }
    incr y
  }
  graf_refresh
}

# ########################################################################
# Event: Button
proc clicked_button { x y } {
  global cube level line
  global buttons busy_flag img

  if { ! $busy_flag } then {
    set busy_flag 1
    set cube(x) $x
    set cube(y) $y
    set line(1) [ lindex $level(1) $cube(y) ]
    if { [ string equal [${buttons}.but($x,$y) cget -image] [ lindex $img 0 ] ] } then {
      ${buttons}.but($x,$y) configure -image [ lindex $img 1 ]
      cube_set_led $x $y $cube(z) 1
#puts "clicked_button $x $y -> $img 1"
    } else {
      ${buttons}.but($x,$y) configure -image [ lindex $img 0 ]
      cube_set_led $x $y $cube(z) 0
#puts "clicked_button $x $y -> $img 0"
    }
    set line(1) [ lindex $level(1) $cube(y) ]
    set busy_flag 0
  }
}

# ############################################################################
proc script_call_proc { l } {
  global workdir cube gvar

  puts [ lindex [split $l] 0 ]
  set todo [ lindex $l 0 ]
  switch $todo {
    "after" {
      if { [ llength $l ] == 2 } then {
        set x [ lindex $l 1 ]
        if { $x < 100 } then { set x 100 }
        if { $x > 10000 } then { set x 10000 }
        after $x
      }
    }
    "cube_set_led" {
      if { [ llength $l ] == 5 } then {
        set x [ lindex $l 1 ]
        set y [ lindex $l 2 ]
        set z [ lindex $l 3 ]
        set led [ lindex $l 4 ]
        if { $x < 0 } then { set x 0 }
        if { $x > $cube(xmax) } then { set x $cube(xmax) }
        if { $y < 0 } then { set y 0 }
        if { $y > $cube(ymax) } then { set y $cube(ymax) }
        if { $z < 0 } then { set z 0 }
        if { $z > $cube(zmax) } then { set z $cube(zmax) }
        if { $led < 0 } then { set led 0 }
        if { $led > $cube(ledmax) } then { set led $cube(ledmax) }
        cube_set_led $x $y $z $led
      }
    }
    "cube_level_0" {
      cube_level_0
    }
    "cube_level_1" {
      cube_level_1
    }
    "cube_load_file" {
      if { [ llength $l ] == 2 } then {
        set filename [ lindex $l 1 ]
        cube_load_file "$workdir/$filename"
      }
    }
    "cube(action)" {
      if { [ llength $l ] == 2 } then {
        set x [ lindex $l 1 ]
        if { $x != 0 } then { set x 1 }
        set cube(action) $x
      }
    }
    "cube(rotate)" {
      if { [ llength $l ] == 2 } then {
        set x [ lindex $l 1 ]
        if { $x != 0 } then { set x 1 }
        set cube(rotate) $x
        cube_switch_rotate
      }
    }
    "cube(x_angle)" {
      if { [ llength $l ] == 2 } then {
        set x [ lindex $l 1 ]
        if { $x < 0 } then { set x 0 }
        if { $x > 90 } then { set x 90 }
        set cube(x_angle) $x
      }
    }
    "cube(y_angle)" {
      if { [ llength $l ] == 2 } then {
        set x [ lindex $l 1 ]
        if { $x < 0 } then { set x 0 }
        if { $x > 90 } then { set x 90 }
        set cube(y_angle) $x
      }
    }
    "cube(z_angle)" {
      if { [ llength $l ] == 2 } then {
        set x [ lindex $l 1 ]
        if { $x < 0 } then { set x 0 }
        if { $x > 90 } then { set x 90 }
        set cube(z_angle) $x
      }
    }
    "cube_trans" {
      if { [ llength $l ] == 4 } then {
        set x [ lindex $l 1 ]
        set y [ lindex $l 2 ]
        set z [ lindex $l 3 ]
        if { $x > 0 } then { set x 1 }
        if { $y > 0 } then { set y 1 }
        if { $z > 0 } then { set z 1 }
        if { $x < 0 } then { set x -1 }
        if { $y < 0 } then { set y -1 }
        if { $z < 0 } then { set z -1 }
        cube_trans $x $y $z
      }
    }
    "cube_send" {
      if { $gvar(led_count) == 0 } then {
        set_xyz $cube(xmax) $cube(ymax) $cube(zmax)
        init_led
      }
      cube_send
    }
    default {
      puts stderr "do not execute: $l"
    }
  }
  update idletask
}

proc script_load_file_and_play { filename } {
  if { [ file exists $filename ] } then {
    if { [ catch {open [file join [file nativename $filename]] r} fid ] } {
      puts stderr "Could not open for reading\n$fid"
    } else {
      set cube(script) {}
      while { ! [ eof $fid ] } {
        set l [ gets $fid ]
        if { [ llength $l ] } then {
#regexp $regexp_vector_4 [join $vector " "] xyz x y z led
          if { [ regexp {^#[.]*} $l ] } then {
            puts $l
          } else {
            lappend cube(script) [ split $l ]
          }
        }
      }
      close $fid
    }
    foreach l $cube(script) {
      script_call_proc $l
    }
  } else {
    puts stderr "Could not find file for reading\n$filename"
  }
}

proc script_load_and_play { } {
  global workdir

  set types {
    { { Led-Cube Files } {.led_script} }
    { { Text Files }     {.txt}      }
    { { All Files }      *           }
  }
  set filename [tk_getOpenFile -parent . -title "Select a file to load" \
     -initialdir "$workdir" -filetypes $types ]
  script_load_file_and_play $filename
}

# ############################################################################
proc init { } {
  global workdir blocksize img

  ini_read

  lappend img [ image create photo -file "$workdir/img/led0_${blocksize}.ppm" ]
  lappend img [ image create photo -file "$workdir/img/led1_${blocksize}.ppm" ]
  lappend img [ image create photo -file "$workdir/img/led_arrow_l_${blocksize}.ppm" ]
  lappend img [ image create photo -file "$workdir/img/led_arrow_r_${blocksize}.ppm" ]
  lappend img [ image create photo -file "$workdir/img/led_arrow_u_${blocksize}.ppm" ]
  lappend img [ image create photo -file "$workdir/img/led_arrow_d_${blocksize}.ppm" ]
  lappend img [ image create photo -file "$workdir/img/led_level1_${blocksize}.ppm" ]
  lappend img [ image create photo -file "$workdir/img/led_level0_${blocksize}.ppm" ]
  lappend img [ image create photo -file "$workdir/img/led_cross_${blocksize}.ppm" ]
  lappend img [ image create photo -file "$workdir/img/led_turn_${blocksize}.ppm" ]
  lappend img [ image create photo -file "$workdir/img/led_cube_${blocksize}.ppm" ]
  lappend img [ image create photo -file "$workdir/img/led_level_${blocksize}.ppm" ]
  lappend img [ image create photo -file "$workdir/img/led_separator_${blocksize}.ppm" ]
  lappend img [ image create photo -file "$workdir/img/led_send_${blocksize}.ppm" ]
}

proc start { } {
  global blocksize cube buttons img

  if { ($cube(xmax)!=$cube(xmax_old)) \
    || ($cube(ymax)!=$cube(ymax_old)) \
    || ($cube(zmax)!=$cube(zmax_old)) } then {
    for { set x 0 } { $x < $cube(xmax_old) } { incr x 1 } {
      for { set y 0 } { $y < $cube(ymax_old) } { incr y 1 } {
        pack forget ${buttons}.but($x,$y)
        destroy ${buttons}.but($x,$y)
      }
    }
    set buttons {}
    set cube(xmax_old) $cube(xmax)
    set cube(ymax_old) $cube(ymax)
    set cube(zmax_old) $cube(zmax)
  }

  .fr.control.can2.fr.1.cube_z configure -from $cube(zmax)
  set cube(z) [ expr $cube(zmax) - 1 ]
  cube_set_1

  if { ! [ llength $buttons ] } then {
    set buttons ".fr.control.can1"
    for { set y 0 } { $y < $cube(ymax) } { incr y 1 } {
      for { set x 0 } { $x < $cube(xmax) } { incr x 1 } {
        button ${buttons}.but($x,$y) -text {} -bitmap {} -image [ lindex $img 0 ] \
          -borderwidth 0 -padx 0 -pady 0 \
          -command "focus ${buttons}.but($x,$y); update idletask; after 50 {clicked_button $x $y}" \
          -state normal -fg white -bg white -disabledforeground "#000000" -highlightcolor "#ff0000"
        place ${buttons}.but($x,$y) -in $buttons \
          -x [ expr 2+($blocksize+4)*($x+1) ] -y [ expr 2+($blocksize+4)*($cube(ymax)-1-$y+1) ]
      }
    }
  }

  set width1 [ expr (1.5*$blocksize)*($cube(xmax)+3) ]
  set height1 [ expr (1.5*$blocksize)*($cube(ymax)+3) ]
  graf_size $width1 $height1

  set width1 [ expr ($blocksize+4)*($cube(xmax)+2) ]
  set height1 [ expr ($blocksize+4)*($cube(ymax)+2) ]
  ${buttons} configure -width $width1 -height $height1

  focus_move 0 0 0 1
  focus ${buttons}.but(0,0)
}

proc end { } {
  ini_write
  exit
}

# ############################################################################
proc ini_write { } {
  global inifilename workdir blocksize cube level

  if { [ catch {open [file join [file nativename "$workdir/$inifilename"]] w} fid ] } {
    puts stderr "Could not open for writing\n$fid"
  } else {
#puts " ini_write blocksize ${blocksize}"
    puts $fid [ join [list "blocksize" $blocksize] {;} ]
#puts " ini_write cube(max) $cube(max)"
    puts $fid [ join [list "cube(max)" [join $cube(max) {;} ]] {;} ]
    puts $fid [ join [list "cube(x_angle)" $cube(x_angle)] {;} ]
    puts $fid [ join [list "cube(y_angle)" $cube(y_angle)] {;} ]
    puts $fid [ join [list "cube(z_angle)" $cube(z_angle)] {;} ]
    puts $fid "cube(1)"
    puts $fid $cube(1)
    puts $fid "level(2)"
    puts $fid $level(2)
    close $fid
  }
}

proc ini_read { } {
  global inifilename workdir blocksize cube level line

  proc test_par { lin par { len 2 } } {
    return [ expr ( [ llength $lin ] == $len ) && ( ! [ string compare $par [lindex $lin 0] ] ) ]
  }

  if { [ catch {open [file join [file nativename "$workdir/$inifilename"]] r} fid ] } {
    puts stderr "Could not open for reading\n$fid"
  } else {
    while { ! [ eof $fid ] } {
      set l [ split [gets $fid] {;} ]
      if { [ llength $l ] > 0 } then {
        set todo [ lindex $l 0 ]
        switch $todo {
          "blocksize" {
            if { [ llength $l ] == 2 } then {
              set x [ expr round( [lindex $l 1] / 10 ) * 10 ]
              if { $x < 20 } then { set x 20 }
              if { $x > 40 } then { set x 40 }
              set blocksize $x
            }
          }
          "cube(x_angle)" {
            if { [ llength $l ] == 2 } then {
              set x [ expr round( [ lindex $l 1 ] / 5 ) * 5 ]
              if { $x < 0 } then { set x 0 }
              if { $x > 90 } then { set x 90 }
              set cube(x_angle) $x
            }
          }
          "cube(y_angle)" {
            if { [ llength $l ] == 2 } then {
              set x [ expr round( [ lindex $l 1 ] / 5 ) * 5 ]
              if { $x < 0 } then { set x 0 }
              if { $x > 90 } then { set x 90 }
              set cube(y_angle) $x
            }
          }
          "cube(z_angle)" {
            if { [ llength $l ] == 2 } then {
              set x [ expr round( [ lindex $l 1 ] / 5 ) * 5 ]
              if { $x < 0 } then { set x 0 }
              if { $x > 90 } then { set x 90 }
              set cube(z_angle) $x
            }
          }
          "cube(max)" {
            if { [ llength $l ] == 4 } then {
              set x [ lindex $l 1 ]
              set y [ lindex $l 2 ]
              set z [ lindex $l 3 ]
              if { $x < 2 } then { set x 2 }
              if { $x > 12 } then { set x 12 }
              if { $y < 2 } then { set y 2 }
              if { $y > 12 } then { set y 12 }
              if { $z < 2 } then { set z 2 }
              if { $z > 12 } then { set z 12 }
              cube_init [ list $x $y $z ]
            }
          }
          "cube(1)" {
            if { [ llength $l ] == 1 } then {
              if { ! [ eof $fid ] } then {
                if { [ info exists "cube(1)" ] } then {
                  set cube(1) [ gets $fid ]
                }
              }
            }
          }
          "level(2)" {
            if { [ llength $l ] == 1 } then {
              if { ! [ eof $fid ] } then {
                if { [ info exists "level(2)" ] } then {
                  set level(2) [ gets $fid ]
                }
              }
            }
          }
          default {
            puts stderr "do not execute: $l"
          }
        }
      }
    }
    close $fid
  }
}

init

# ########################################################################
wm title . "LED Cube Editor"
frame .fr

graf_create .fr.screen
frame .fr.control
canvas .fr.control.can1
canvas .fr.control.can2

frame .fr.menubar -relief raised -bd 2
button .fr.menubar.quit -text "Ende" -command "end" -state normal
button .fr.menubar.save -text "Speichern" -command "cube_save" -state normal
button .fr.menubar.load -text "Laden" -command "cube_load" -state normal
button .fr.menubar.loadscript -text "Script" -command "script_load_and_play" -state normal

button .fr.menubar.l -image [ lindex $img 2 ] -command "cube_trans -1 0 0" -state normal
button .fr.menubar.r -image [ lindex $img 3 ] -command "cube_trans +1 0 0" -state normal
button .fr.menubar.u -image [ lindex $img 4 ] -command "cube_trans 0 +1 0" -state normal
button .fr.menubar.d -image [ lindex $img 5 ] -command "cube_trans 0 -1 0" -state normal
button .fr.menubar.level_1 -image [ lindex $img 6 ] -command "cube_level_1" -state normal
button .fr.menubar.level_0 -image [ lindex $img 7 ] -command "cube_level_0" -state normal
checkbutton .fr.menubar.cube_level -variable cube(action) \
  -selectimage [ lindex $img 10 ] -image [ lindex $img 11 ] \
  -onvalue 1 -offvalue 0 -state normal -indicatoron 0
checkbutton .fr.menubar.turn_move -variable cube(rotate) \
  -selectimage [ lindex $img 9 ] -image [ lindex $img 8 ] \
  -command "cube_switch_rotate" \
  -onvalue 1 -offvalue 0 -state normal -indicatoron 0
button .fr.menubar.sep0 -image [ lindex $img 12 ] -state disabled -relief flat -border 0
button .fr.menubar.sep1 -image [ lindex $img 12 ] -state disabled -relief flat -border 0
button .fr.menubar.sep2 -image [ lindex $img 12 ] -state disabled -relief flat -border 0
button .fr.menubar.sep3 -image [ lindex $img 12 ] -state disabled -relief flat -border 0
button .fr.menubar.sep4 -image [ lindex $img 12 ] -state disabled -relief flat -border 0
button .fr.menubar.sep5 -image [ lindex $img 12 ] -state disabled -relief flat -border 0
button .fr.menubar.send -image [ lindex $img 13 ] -command "cube_send" -state normal

pack .fr.menubar.quit \
  .fr.menubar.sep0 .fr.menubar.loadscript .fr.menubar.send \
  .fr.menubar.sep1 .fr.menubar.save .fr.menubar.load \
  .fr.menubar.sep2 .fr.menubar.d .fr.menubar.u \
  .fr.menubar.sep3 .fr.menubar.r .fr.menubar.l \
  .fr.menubar.sep4 .fr.menubar.level_0 .fr.menubar.level_1 \
  .fr.menubar.sep5 .fr.menubar.turn_move .fr.menubar.cube_level \
  -side right -fill y -expand 0
pack .fr.menubar -side top -fill x -expand 0
tk_menuBar .fr.menubar

frame .fr.control.can2.fr
frame .fr.control.can2.fr.1
frame .fr.control.can2.fr.2
scale .fr.control.can2.fr.1.cube_z -label "Z - Ebene" -orient vertical -length 150 -width 15 \
  -sliderlength 20 -sliderrelief raised -state normal \
  -from $cube(zmax) -to -1 -resolution 1 -tickinterval 0 \
  -showvalue 1 -digits 1 -bigincrement 1 -borderwidth 1 \
  -variable cube(z) -command "buttons_reset; focus_move 0 0 0 "
scale .fr.control.can2.fr.2.x_angle -label "X - Winkel" -orient horizontal -length 150 -width 15 \
  -sliderlength 20 -sliderrelief raised -state normal \
  -from 0 -to 90 -resolution 5 -tickinterval 15 \
  -showvalue 0 -digits 1 -bigincrement 15 -borderwidth 1 \
  -variable cube(x_angle)
scale .fr.control.can2.fr.2.y_angle -label "Y - Winkel" -orient horizontal -length 150 -width 15 \
  -sliderlength 20 -sliderrelief raised -state normal \
  -from 0 -to 90 -resolution 5 -tickinterval 15 \
  -showvalue 0 -digits 1 -bigincrement 15 -borderwidth 1 \
  -variable cube(y_angle)
scale .fr.control.can2.fr.2.z_angle -label "Z - Winkel" -orient horizontal -length 150 -width 15 \
  -sliderlength 20 -sliderrelief raised -state normal \
  -from 0 -to 90 -resolution 5 -tickinterval 15 \
  -showvalue 0 -digits 1 -bigincrement 15 -borderwidth 1 \
  -variable cube(z_angle)

pack .fr.control.can2.fr.1.cube_z -side top -fill none -expand 1
pack .fr.control.can2.fr.2.x_angle .fr.control.can2.fr.2.y_angle .fr.control.can2.fr.2.z_angle -side top -fill none -expand 1
pack .fr.control.can2.fr.1 .fr.control.can2.fr.2 -side left -fill none -expand 1
pack .fr.control.can2.fr -side bottom -fill none -expand 1
pack .fr.control.can1 -side top -fill none -expand 0
pack .fr.control.can2 -side bottom -fill both -expand 0
pack .fr.screen -side left -fill none -expand 1
pack .fr.control -side right -fill none -expand 1
pack .fr -padx 2 -pady 2 -fill both -expand 1

# ########################################################################
bind . <KeyRelease-Left>  { focus_move -1 0 0 }
bind . <KeyRelease-Right> { focus_move +1 0 0 }
bind . <KeyRelease-Down>  { focus_move 0 -1 0 }
bind . <KeyRelease-Up>    { focus_move 0 +1 0 }
bind . <KeyRelease-Next>  { focus_move 0 0 -1 }
bind . <KeyRelease-Prior> { focus_move 0 0 +1 }
bind . <Shift-Left>  { cube_trans -1 0 0; focus_move +1 0 0 }
bind . <Shift-Right> { cube_trans +1 0 0; focus_move -1 0 0 }
bind . <Shift-Up>    { cube_trans 0 -1 0; focus_move 0 +1 0 }
bind . <Shift-Down>  { cube_trans 0 +1 0; focus_move 0 -1 0 }
#bind . <Control-Shift-Left>  { cube_move -1 0 0 }
#bind . <Control-Shift-Right> { cube_move +1 0 0 }
#bind . <Control-Shift-Down>  { cube_move 0 -1 0 }
#bind . <Control-Shift-Up>    { cube_move 0 +1 0 }
#bind . <Control-Shift-Next>  { cube_move 0 0 -1 }
#bind . <Control-Shift-Prior> { cube_move 0 0 +1 }
bind . <KeyRelease-Return> { global cube; clicked_button $cube(x) $cube(y) }
bind . <Control-c>  { cube_level_copy }
bind . <Control-x>  { cube_level_cut }
bind . <Control-v>  { cube_level_paste }
bind . <Control-Insert>  { cube_level_copy }
bind . <Shift-Delete>  { cube_level_cut }
bind . <Shift-Insert>  { cube_level_paste }
bind . <Delete>  { cube_level_0 }
bind . <Insert>  { cube_level_1 }
bind . <Control-s>  { cube_send }

bind . <Control-G>  { cube_generate_3dv }

#bind . <Any-KeyPress> {puts "Name %K"}

# ########################################################################

start
