global regexp_vector
set regexp_vector {^([+-\.0-9A-Za-z]*)[ ]*([+-\.0-9A-Za-z]*)[ ]*([+-\.0-9A-Za-z]*)$}

# Transformationsmatrix
proc pi {} {
  return 3.141592654
}

proc matrix_unit { } {
  return {{0.0 0.0 0.0} {1.0 0.0 0.0} {0.0 1.0 0.0} {0.0 0.0 1.0}}
}

proc matrix_set { matrix } {
  return $matrix
}

proc matrix_scale { matrix a b c } {
  set nmatrix {}
  foreach { x y z } $matrix {
    lappend nmatrix [ list [expr $x * $a] [expr $y * $b] [expr $z * $c] ]
  }
  return $nmatrix
}

proc matrix_move { matrix x y z } {
  global regexp_vector
  set vector [lindex $matrix 0]
  regexp $regexp_vector [join $vector " "] XYZ X0 Y0 Z0
  return [ lreplace $matrix 0 0 \
    [list [expr $X0 + $x] [expr $Y0 + $y] [expr $Z0 + $z]] ]
}

proc matrix_x_rotate { matrix phi } {
  global regexp_vector
#puts "matrix_x_rotate $phi $matrix"

  set phi [expr $phi*([pi] / 180.0)]
  set cos_phi [expr cos($phi)]
  set sin_phi [expr sin($phi)]

  regexp $regexp_vector [join [lindex $matrix 0] " "] XYZ X0 Y0 Z0
  regexp $regexp_vector [join [lindex $matrix 1] " "] XYZ XX YX ZX
  regexp $regexp_vector [join [lindex $matrix 2] " "] XYZ XY YY ZY
  regexp $regexp_vector [join [lindex $matrix 3] " "] XYZ XZ YZ ZZ

  set Y0_ [expr $Y0 * $cos_phi + $Z0 * $sin_phi]
  set YX_ [expr $YX * $cos_phi + $ZX * $sin_phi]
  set YY_ [expr $YY * $cos_phi + $ZY * $sin_phi]
  set YZ_ [expr $YZ * $cos_phi + $ZZ * $sin_phi]

  set Z0_ [expr $Z0 * $cos_phi - $Y0 * $sin_phi]
  set ZX_ [expr $ZX * $cos_phi - $YX * $sin_phi]
  set ZY_ [expr $ZY * $cos_phi - $YY * $sin_phi]
  set ZZ_ [expr $ZZ * $cos_phi - $YZ * $sin_phi]

  return [ list [list $X0 $Y0_ $Z0_] [list $XX $YX_ $ZX_] \
                   [list $XY $YY_ $ZY_] [list $XZ $YZ_ $ZZ_] ]
}

proc matrix_y_rotate { matrix phi } {
  global regexp_vector
#puts "matrix_y_rotate $phi $matrix"

  set phi [expr $phi*([pi] / 180.0)]
  set cos_phi [expr cos($phi)]
  set sin_phi [expr sin($phi)]

  regexp $regexp_vector [join [lindex $matrix 0] " "] XYZ X0 Y0 Z0
  regexp $regexp_vector [join [lindex $matrix 1] " "] XYZ XX YX ZX
  regexp $regexp_vector [join [lindex $matrix 2] " "] XYZ XY YY ZY
  regexp $regexp_vector [join [lindex $matrix 3] " "] XYZ XZ YZ ZZ

  set X0_ [expr $X0 * $cos_phi + $Z0 * $sin_phi]
  set XX_ [expr $XX * $cos_phi + $ZX * $sin_phi]
  set XY_ [expr $XY * $cos_phi + $ZY * $sin_phi]
  set XZ_ [expr $XZ * $cos_phi + $ZZ * $sin_phi]

  set Z0_ [expr $Z0 * $cos_phi - $X0 * $sin_phi]
  set ZX_ [expr $ZX * $cos_phi - $XX * $sin_phi]
  set ZY_ [expr $ZY * $cos_phi - $XY * $sin_phi]
  set ZZ_ [expr $ZZ * $cos_phi - $XZ * $sin_phi]

  return [ list [list $X0_ $Y0 $Z0_] [list $XX_ $YX $ZX_] \
                   [list $XY_ $YY $ZY_] [list $XZ_ $YZ $ZZ_] ]
}

proc matrix_z_rotate { matrix phi } {
  global regexp_vector
#puts "matrix_z_rotate $phi $matrix"

  set phi [expr $phi*([pi] / 180.0)]
  set cos_phi [expr cos($phi)]
  set sin_phi [expr sin($phi)]

  regexp $regexp_vector [join [lindex $matrix 0] " "] XYZ X0 Y0 Z0
  regexp $regexp_vector [join [lindex $matrix 1] " "] XYZ XX YX ZX
  regexp $regexp_vector [join [lindex $matrix 2] " "] XYZ XY YY ZY
  regexp $regexp_vector [join [lindex $matrix 3] " "] XYZ XZ YZ ZZ

  set Y0_ [expr $Y0 * $cos_phi + $X0 * $sin_phi]
  set YX_ [expr $YX * $cos_phi + $XX * $sin_phi]
  set YY_ [expr $YY * $cos_phi + $XY * $sin_phi]
  set YZ_ [expr $YZ * $cos_phi + $XZ * $sin_phi]

  set X0_ [expr $X0 * $cos_phi - $Y0 * $sin_phi]
  set XX_ [expr $XX * $cos_phi - $YX * $sin_phi]
  set XY_ [expr $XY * $cos_phi - $YY * $sin_phi]
  set XZ_ [expr $XZ * $cos_phi - $YZ * $sin_phi]
  return [ list [list $X0_ $Y0_ $Z0] [list $XX_ $YX_ $ZX] \
                   [list $XY_ $YY_ $ZY] [list $XZ_ $YZ_ $ZZ] ]
}

proc matrix_transform { matrix tmatrix } {
  global regexp_vector

  regexp $regexp_vector [join [lindex $matrix 0] " "] XYZ X0 Y0 Z0
  regexp $regexp_vector [join [lindex $matrix 1] " "] XYZ XX YX ZX
  regexp $regexp_vector [join [lindex $matrix 2] " "] XYZ XY YY ZY
  regexp $regexp_vector [join [lindex $matrix 3] " "] XYZ XZ YZ ZZ

  regexp $regexp_vector [join [lindex $tmatrix 0] " "] XYZ tX0 tY0 tZ0
  regexp $regexp_vector [join [lindex $tmatrix 1] " "] XYZ tXX tYX tZX
  regexp $regexp_vector [join [lindex $tmatrix 2] " "] XYZ tXY tYY tZY
  regexp $regexp_vector [join [lindex $tmatrix 3] " "] XYZ tXZ tYZ tZZ

  set X0_ [expr $X0 * $tXX + $Y0 * $tXY + $Z0 * $tXZ + $tX0]
  set XX_ [expr $XX * $tXX + $YX * $tXY + $ZX * $tXZ]
  set XY_ [expr $XY * $tXX + $YY * $tXY + $ZY * $tXZ]
  set XZ_ [expr $XZ * $tXX + $YZ * $tXY + $ZZ * $tXZ]

  set Y0_ [expr $X0 * $tYX + $Y0 * $tYY + $Z0 * $tYZ + $tY0]
  set YX_ [expr $XX * $tYX + $YX * $tYY + $ZX * $tYZ]
  set YY_ [expr $XY * $tYX + $YY * $tYY + $ZY * $tYZ]
  set YZ_ [expr $XZ * $tYX + $YZ * $tYY + $ZZ * $tYZ]

  set Z0_ [expr $X0 * $tZX + $Y0 * $tZY + $Z0 * $tZZ + $tZ0]
  set ZX_ [expr $XX * $tZX + $YX * $tZY + $ZX * $tZZ]
  set ZY_ [expr $XY * $tZX + $YY * $tZY + $ZY * $tZZ]
  set ZZ_ [expr $XZ * $tZX + $YZ * $tZY + $ZZ * $tZZ]

  return [ list [list $X0_ $Y0_ $Z0_] [list $XX_ $YX_ $ZX_] \
                   [list $XY_ $YY_ $ZY_] [list $XZ_ $YZ_ $ZZ_] ]
}

proc vector_value { vector } {
  global regexp_vector

  regexp $regexp_vector [join $vector " "] XYZ X Y Z
  return [ expr sqrt( $X * $X + $Y * $Y + $Z * $Z ) ]
}

proc vector_diff { vector1 vector2 } {
  global regexp_vector

  regexp $regexp_vector [join $vector1 " "] XYZ X1 Y1 Z1
  regexp $regexp_vector [join $vector2 " "] XYZ X2 Y2 Z2
  return [ list \
    [expr $X2 - $X1] [expr $Y2 - $Y1] [expr $Z2 - $Z1] ]
}

proc vector_scalarproduct { vector1 vector2 } {
  global regexp_vector

  regexp $regexp_vector [join $vector1 " "] XYZ X1 Y1 Z1
  regexp $regexp_vector [join $vector2 " "] XYZ X2 Y2 Z2
  return [ expr $X1 * $X2 + $Y1 * $Y2 + $Z1 * $Z2 ]
}

proc vector_angle { vector1 vector2 } {
  set v1b [ vector_value $vector1 ]
  set v2b [ vector_value $vector2 ]
  if { ($v1b == 0.0) || ($v2b == 0.0) } then {
    return 1.0
  } else {
    return [ expr acos( [ vector_scalarproduct $vector1 $vector2 ] / $v1b / $v2b ) * 180.0 / [pi] ]
  }
}

proc vectors_bounds { vectors bounds } {
  global regexp_vector

  if { [ llength $vectors ] } then {
    set vector [ lindex $vectors 0 ]
    regexp $regexp_vector [join $vector " "] XYZ XMIN YMIN ZMIN
    regexp $regexp_vector [join $vector " "] XYZ XMAX YMAX ZMAX
    foreach vector $vectors {
      regexp $regexp_vector [join $vector " "] XYZ X Y Z
      set XMIN [ expr ($X < $XMIN) ? $X : $XMIN ]
      set YMIN [ expr ($Y < $YMIN) ? $Y : $YMIN ]
      set ZMIN [ expr ($Z < $ZMIN) ? $Z : $ZMIN ]
      set XMAX [ expr ($X < $XMAX) ? $X : $XMAX ]
      set YMAX [ expr ($Y < $YMAX) ? $Y : $YMAX ]
      set ZMAX [ expr ($Z < $ZMAX) ? $Z : $ZMAX ]
    }
    return [ list [list $XMIN $YMIN $ZMIN] [list $XMAX $YMAX $ZMAX] ]
  } else {
    return $bounds
  }
}

proc vectors_transform { vectors matrix } {
  global regexp_vector

  regexp $regexp_vector [join [lindex $matrix 0] " "] XYZ X0 Y0 Z0
  regexp $regexp_vector [join [lindex $matrix 1] " "] XYZ XX YX ZX
  regexp $regexp_vector [join [lindex $matrix 2] " "] XYZ XY YY ZY
  regexp $regexp_vector [join [lindex $matrix 3] " "] XYZ XZ YZ ZZ

  set new_vectors {}
  foreach vector $vectors {
    regexp $regexp_vector [join $vector " "] XYZ X Y Z
    lappend new_vectors [ list \
      [expr $X0 + $X * $XX + $Y * $XY + $Z * $XZ] \
      [expr $Y0 + $X * $YX + $Y * $YY + $Z * $YZ] \
      [expr $Z0 + $X * $ZX + $Y * $ZY + $Z * $ZZ] ]
  }
  return $new_vectors
}
